@@ -4,7 +4,7 @@
  * This file is part of libaccounts-glib
  *
  * Copyright (C) 2009-2010 Nokia Corporation.
- * Copyright (C) 2012 Canonical Ltd.
+ * Copyright (C) 2012-2013 Canonical Ltd.
  * Copyright (C) 2012 Intel Corporation.
  *
  * Contact: Alberto Mardegan <alberto.mardegan@canonical.com>
@@ -54,15 +54,25 @@
 #include "ag-internals.h"
 #include "ag-service.h"
 #include "ag-util.h"
+#include <errno.h>
+#include <fcntl.h>
 #include <sched.h>
 #include <sqlite3.h>
 #include <string.h>
+#include <sys/stat.h>
+#include <sys/types.h>
 #include <unistd.h>
 
 #ifndef DATABASE_DIR
 #define DATABASE_DIR "libaccounts-glib"
 #endif
 
+#ifdef DISABLE_WAL
+#define JOURNAL_MODE "TRUNCATE"
+#else
+#define JOURNAL_MODE "WAL"
+#endif
+
 enum
 {
     PROP_0,
@@ -603,12 +613,12 @@ dbus_filter_callback (G_GNUC_UNUSED GDBusConnection *dbus_conn,
          * created or deleted from another instance.
          * We must emit the signals, and cache the newly created account for a
          * while, because the application is likely to inspect it */
-        account = g_object_new (AG_TYPE_ACCOUNT,
-                                "manager", manager,
-                                "provider", provider_name,
-                                "id", account_id,
-                                "foreign", created,
-                                NULL);
+        account = g_initable_new (AG_TYPE_ACCOUNT, NULL, NULL,
+                                  "manager", manager,
+                                  "provider", provider_name,
+                                  "id", account_id,
+                                  "foreign", created,
+                                  NULL);
         g_return_if_fail (AG_IS_ACCOUNT (account));
 
         g_object_weak_ref (G_OBJECT (account), account_weak_notify, manager);
@@ -1032,10 +1042,11 @@ setup_db_options (sqlite3 *db)
     }
 
     error = NULL;
-    ret = sqlite3_exec (db, "PRAGMA journal_mode = WAL", NULL, NULL, &error);
+    ret = sqlite3_exec (db, "PRAGMA journal_mode = " JOURNAL_MODE, NULL, NULL,
+                        &error);
     if (ret != SQLITE_OK)
     {
-        g_warning ("%s: couldn't set journal mode to WAL (%s)",
+        g_warning ("%s: couldn't set journal mode to " JOURNAL_MODE " (%s)",
                    G_STRFUNC, error);
         sqlite3_free (error);
     }
@@ -1133,6 +1144,30 @@ create_db (sqlite3 *db)
     return TRUE;
 }
 
+static inline gboolean
+file_is_read_only (const gchar *filename)
+{
+    int fd;
+
+    /* FIXME: Here we could just use access(); however, because of bug
+     * https://bugs.launchpad.net/bugs/1220713, if the access is blocked by
+     * apparmor the access() call would still report that the file is
+     * writeable.
+     */
+    fd = open (filename, O_RDWR);
+    if (fd == -1)
+    {
+        if (errno == EACCES || errno == EROFS)
+            return TRUE;
+    }
+    else
+    {
+        close (fd);
+    }
+
+    return FALSE;
+}
+
 static gboolean
 open_db (AgManager *manager)
 {
@@ -1141,7 +1176,7 @@ open_db (AgManager *manager)
     gchar *filename, *pathname;
     gint version;
     gboolean ok = TRUE;
-    int ret;
+    int ret, flags;
 
     basedir = g_getenv ("ACCOUNTS");
     if (G_LIKELY (!basedir))
@@ -1158,7 +1193,18 @@ open_db (AgManager *manager)
     {
         filename = g_build_filename (basedir, "accounts.db", NULL);
     }
-    ret = sqlite3_open (filename, &priv->db);
+
+    /* First, check if the file exists and is writeable */
+    if (file_is_read_only (filename))
+    {
+        DEBUG_INFO ("Opening DB in read-only mode");
+        flags = SQLITE_OPEN_READONLY;
+    }
+    else
+    {
+        flags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE;
+    }
+    ret = sqlite3_open_v2 (filename, &priv->db, flags, NULL);
     g_free (filename);
 
     if (ret != SQLITE_OK)
@@ -1859,23 +1905,17 @@ ag_manager_load_account (AgManager *manager, AgAccountId account_id,
         return g_object_ref (account);
 
     /* the account is not loaded; do it now */
-    account = g_object_new (AG_TYPE_ACCOUNT,
-                            "manager", manager,
-                            "id", account_id,
-                            NULL);
+    account = g_initable_new (AG_TYPE_ACCOUNT, NULL, error,
+                              "manager", manager,
+                              "id", account_id,
+                              NULL);
     if (G_LIKELY (account))
     {
         g_object_weak_ref (G_OBJECT (account), account_weak_notify, manager);
         g_hash_table_insert (priv->accounts, GUINT_TO_POINTER (account_id),
                              account);
     }
-    else if (priv->last_error != NULL)
-    {
-        g_set_error_literal (error,
-                             priv->last_error->domain,
-                             priv->last_error->code,
-                             priv->last_error->message);
-    }
+
     return account;
 }
 
@@ -1898,10 +1938,10 @@ ag_manager_create_account (AgManager *manager, const gchar *provider_name)
 
     g_return_val_if_fail (AG_IS_MANAGER (manager), NULL);
 
-    account = g_object_new (AG_TYPE_ACCOUNT,
-                            "manager", manager,
-                            "provider", provider_name,
-                            NULL);
+    account = g_initable_new (AG_TYPE_ACCOUNT, NULL, NULL,
+                              "manager", manager,
+                              "provider", provider_name,
+                              NULL);
     return account;
 }
 
@@ -2084,6 +2124,16 @@ ag_manager_list_services_by_type (AgManager *manager, const gchar *service_type)
     return services;
 }
 
+static GError *
+sqlite_error_to_gerror (int db_error, sqlite3 *db)
+{
+    AgAccountsError code = (db_error == SQLITE_READONLY) ?
+        AG_ACCOUNTS_ERROR_READONLY : AG_ACCOUNTS_ERROR_DB;
+    return g_error_new (AG_ACCOUNTS_ERROR, code,
+                        "Got error: %s (%d)",
+                        sqlite3_errmsg (db), db_error);
+}
+
 void
 _ag_manager_exec_transaction (AgManager *manager, const gchar *sql,
                               AgAccountChanges *changes, AgAccount *account,
@@ -2097,9 +2147,7 @@ _ag_manager_exec_transaction (AgManager *manager, const gchar *sql,
     ret = prepare_transaction_statements (priv);
     if (G_UNLIKELY (ret != SQLITE_OK))
     {
-        error = g_error_new (AG_ACCOUNTS_ERROR, AG_ACCOUNTS_ERROR_DB,
-                             "Got error: %s (%d)",
-                             sqlite3_errmsg (priv->db), ret);
+        error = sqlite_error_to_gerror (ret, priv->db);
         goto finish;
     }
 
@@ -2122,9 +2170,7 @@ _ag_manager_exec_transaction (AgManager *manager, const gchar *sql,
 
     if (ret != SQLITE_DONE)
     {
-        error = g_error_new (AG_ACCOUNTS_ERROR, AG_ACCOUNTS_ERROR_DB,
-                             "Got error: %s (%d)",
-                             sqlite3_errmsg (priv->db), ret);
+        error = sqlite_error_to_gerror (ret, priv->db);
         goto finish;
     }
 
@@ -2152,9 +2198,7 @@ _ag_manager_exec_transaction_blocking (AgManager *manager, const gchar *sql,
     ret = prepare_transaction_statements (priv);
     if (G_UNLIKELY (ret != SQLITE_OK))
     {
-        *error = g_error_new (AG_ACCOUNTS_ERROR, AG_ACCOUNTS_ERROR_DB,
-                              "Got error: %s (%d)",
-                              sqlite3_errmsg (priv->db), ret);
+        *error = sqlite_error_to_gerror (ret, priv->db);
         return;
     }
 
@@ -2177,9 +2221,7 @@ _ag_manager_exec_transaction_blocking (AgManager *manager, const gchar *sql,
 
     if (ret != SQLITE_DONE)
     {
-        *error = g_error_new (AG_ACCOUNTS_ERROR, AG_ACCOUNTS_ERROR_DB,
-                              "Got error: %s (%d)",
-                              sqlite3_errmsg (priv->db), ret);
+        *error = sqlite_error_to_gerror (ret, priv->db);
         return;
     }
 
